(globalThis.TURBOPACK||(globalThis.TURBOPACK=[])).push(["object"==typeof document?document.currentScript:void 0,613952,e=>{"use strict";var r=e.i(640865),o=e.i(690537),l=e.i(774923),t=e.i(896772),s=e.i(178675);function a(){let e=(0,r.useTheme)(),a=(0,l.useRecceThemeOptional)(),c=(0,s.useIsDark)(),[u,n]=(0,o.useState)(!1);(0,o.useEffect)(()=>{n(!0)},[]);let i=!!u&&(a?"dark"===a.resolvedMode:c);return{isDark:i,theme:e,background:{default:i?t.colors.neutral[900]:t.colors.white,paper:i?t.colors.neutral[800]:t.colors.white,subtle:i?t.colors.neutral[800]:t.colors.neutral[50],emphasized:i?t.colors.neutral[700]:t.colors.neutral[100]},text:{primary:i?t.colors.neutral[50]:t.colors.neutral[900],secondary:i?t.colors.neutral[400]:t.colors.neutral[600],disabled:i?t.colors.neutral[500]:t.colors.neutral[400],inverted:i?t.colors.neutral[900]:t.colors.neutral[50]},border:{light:i?t.colors.neutral[700]:t.colors.neutral[200],default:i?t.colors.neutral[600]:t.colors.neutral[300],strong:i?t.colors.neutral[500]:t.colors.neutral[400]},status:{added:{bg:i?t.colors.green[900]:t.colors.green[100],text:i?t.colors.neutral[50]:t.colors.neutral[900]},removed:{bg:i?t.colors.red[950]:t.colors.red[200],text:i?t.colors.neutral[50]:t.colors.neutral[900]},modified:{bg:i?t.colors.yellow[900]:t.colors.amber[100],text:i?t.colors.neutral[50]:t.colors.neutral[900]}},interactive:{hover:i?t.colors.neutral[700]:t.colors.neutral[100],active:i?t.colors.neutral[600]:t.colors.neutral[200],focus:t.colors.iochmara[500]}}}e.s(["useThemeColors",()=>a])},418382,e=>{"use strict";var r=e.i(10049),o=e.i(310402),l=e.i(664745);function t(){return(0,r.jsx)(o.ReactFlowProvider,{children:(0,r.jsx)(l.LineageViewOss,{interactive:!0})})}e.s(["LineagePageOss",()=>t])},670337,e=>{"use strict";var r=e.i(10049),o=e.i(418382);function l(){return(0,r.jsx)(o.LineagePageOss,{})}e.s(["default",()=>l])}]);