.ag-body-viewport,.ag-center-cols-viewport{background-color:var(--ag-background-color)}.ag-header{z-index:1;position:sticky;top:0}.ag-header-cell-comp-wrapper{align-items:center;width:100%;display:flex}.ag-header-cell-label{flex:1;align-items:center;display:flex;overflow:hidden}.ag-header-cell-label .ag-sort-indicator-container{margin-left:4px}.ag-header-cell-menu-button{margin-left:auto}.ag-header-icon{flex-shrink:0}.grid-header:hover .pin-icon,.grid-header:hover .key-icon{display:block}.ag-row:hover .row-context-menu{visibility:visible;width:auto}.cell-show-context-menu{overflow:visible}
:root,.light{--schema-color-added:#ccffd8;--schema-color-added-hover:#c1f6cd;--schema-color-added-active:#a6f2b8;--schema-color-removed:#ffd7d5;--schema-color-removed-hover:#f6c3c1;--schema-color-removed-active:#f2aaa6;--schema-color-normal:#fff;--schema-color-normal-hover:#f1f1f1;--schema-color-normal-active:#e6e6e6;--schema-color-reordered:#ffad15;--schema-color-reordered-hover:#f9a406;--schema-color-reordered-active:#e09306}.dark{--schema-color-added:#1a4d1a;--schema-color-added-hover:#226122;--schema-color-added-active:#2a752a;--schema-color-removed:#5c1f1f;--schema-color-removed-hover:#702929;--schema-color-removed-active:#843333;--schema-color-normal:#262626;--schema-color-normal-hover:#404040;--schema-color-normal-active:#525252;--schema-color-reordered:#b45309;--schema-color-reordered-hover:#92400e;--schema-color-reordered-active:#78350f}.row-added{background-color:var(--schema-color-added)}.row-added:hover,.row-added[aria-selected=true],.row-added.ag-row-selected{background-color:var(--schema-color-added-hover)}.row-added[aria-selected=true]:hover,.row-added.ag-row-selected:hover{background-color:var(--schema-color-added-active)}.row-removed{background-color:var(--schema-color-removed)}.row-removed:hover,.row-removed[aria-selected=true],.row-removed.ag-row-selected{background-color:var(--schema-color-removed-hover)}.row-removed[aria-selected=true]:hover,.row-removed.ag-row-selected:hover{background-color:var(--schema-color-removed-active)}.row-normal{background-color:var(--schema-color-normal)}.row-normal:hover,.row-normal[aria-selected=true],.row-normal.ag-row-selected{background-color:var(--schema-color-normal-hover)}.row-normal[aria-selected=true]:hover,.row-normal.ag-row-selected:hover{background-color:var(--schema-color-normal-active)}.row-selectable{cursor:pointer}.schema-column[aria-selected=true]{outline-offset:0;outline:none}.schema-column-index{text-align:right;filter:brightness(.95)}.type-badge{border-radius:4px;padding:2px 6px;font-size:.75rem;line-height:1.2;display:inline-block}.type-badge-removed{background-color:var(--schema-color-removed);margin-right:4px}.type-badge-added{background-color:var(--schema-color-added)}.column-index-reordered{background-color:var(--schema-color-reordered)}.row-normal:hover .column-index-reordered,.row-normal[aria-selected=true] .column-index-reordered,.row-normal.ag-row-selected .column-index-reordered{background-color:var(--schema-color-reordered-hover)}.row-normal[aria-selected=true]:hover .column-index-reordered,.row-normal.ag-row-selected:hover .column-index-reordered{background-color:var(--schema-color-reordered-active)}
.grid-header:hover .pin-icon,.grid-header:hover .key-icon{display:block}.ag-row:hover .row-context-menu{visibility:visible;width:auto}.cell-show-context-menu{overflow:visible}
