(globalThis.TURBOPACK||(globalThis.TURBOPACK=[])).push(["object"==typeof document?document.currentScript:void 0,613952,e=>{"use strict";var r=e.i(640865),o=e.i(690537),l=e.i(774923),t=e.i(896772),s=e.i(178675);function a(){let e=(0,r.useTheme)(),a=(0,l.useRecceThemeOptional)(),c=(0,s.useIsDark)(),[n,u]=(0,o.useState)(!1);(0,o.useEffect)(()=>{u(!0)},[]);let i=!!n&&(a?"dark"===a.resolvedMode:c);return{isDark:i,theme:e,background:{default:i?t.colors.neutral[900]:t.colors.white,paper:i?t.colors.neutral[800]:t.colors.white,subtle:i?t.colors.neutral[800]:t.colors.neutral[50],emphasized:i?t.colors.neutral[700]:t.colors.neutral[100]},text:{primary:i?t.colors.neutral[50]:t.colors.neutral[900],secondary:i?t.colors.neutral[400]:t.colors.neutral[600],disabled:i?t.colors.neutral[500]:t.colors.neutral[400],inverted:i?t.colors.neutral[900]:t.colors.neutral[50]},border:{light:i?t.colors.neutral[700]:t.colors.neutral[200],default:i?t.colors.neutral[600]:t.colors.neutral[300],strong:i?t.colors.neutral[500]:t.colors.neutral[400]},status:{added:{bg:i?t.colors.green[900]:t.colors.green[100],text:i?t.colors.neutral[50]:t.colors.neutral[900]},removed:{bg:i?t.colors.red[950]:t.colors.red[200],text:i?t.colors.neutral[50]:t.colors.neutral[900]},modified:{bg:i?t.colors.yellow[900]:t.colors.amber[100],text:i?t.colors.neutral[50]:t.colors.neutral[900]}},interactive:{hover:i?t.colors.neutral[700]:t.colors.neutral[100],active:i?t.colors.neutral[600]:t.colors.neutral[200],focus:t.colors.iochmara[500]}}}e.s(["useThemeColors",()=>a])},408821,e=>{"use strict";var r=e.i(10049),o=e.i(613952),l=e.i(998963),t=e.i(631597),s=e.i(131426),a=e.i(677278),c=e.i(690537);function n({error:e,reset:n}){let{background:u,border:i,text:d}=(0,o.useThemeColors)();return(0,c.useEffect)(()=>{a.captureException(e,{tags:{errorBoundary:"app",digest:e.digest},extra:{componentStack:e.stack}}),console.error("App Error Boundary caught error:",e)},[e]),(0,r.jsx)(l.default,{sx:{height:"100%",bgcolor:u.subtle,display:"flex",alignItems:"center",justifyContent:"center"},children:(0,r.jsxs)(l.default,{sx:{p:2,display:"flex",flexDirection:"column",justifyContent:"flex-start",bgcolor:u.paper,border:"solid 1px",borderColor:i.light,borderRadius:"8px",minHeight:"200px",boxShadow:3},children:[(0,r.jsx)(s.default,{variant:"h6",sx:{width:"800px",mb:2},children:"You have encountered an error"}),(0,r.jsx)(l.default,{sx:{flex:1,fontSize:"10pt",color:d.secondary,mb:2},children:e.message||String(e)}),e.digest&&(0,r.jsxs)(l.default,{sx:{fontSize:"9pt",color:d.disabled,mb:2},children:["Error ID: ",e.digest]}),(0,r.jsx)(t.default,{sx:{justifySelf:"center",alignSelf:"center",mt:"20px"},color:"iochmara",variant:"contained",size:"small",onClick:()=>{n()},children:"Try again"})]})})}e.s(["default",()=>n])}]);